﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Windows.Forms;

namespace CustomLearning
{
    public partial class Form1 : Form
    {
        // ノードの数
        const int num_input_nodes = 28 * 28;
        const int num_hidden_nodes = 100;
        int num_output_nodes;
        // 学習率
        double alpha = 0.1;

        int num_training_data;
        int num_test_data;

        NeuralNetwork nn;
        double[] currentPixel = new double[num_input_nodes];
        int currentLabelIndex;

        List<double[]> pixel = new List<double[]>();
        List<double[]> label = new List<double[]>();
        List<int> labelIndex = new List<int>();
        string[] labelName;
        bool modifyFlag = false;

        int px0;
        int py0;

        Random random = new Random();

        public Form1()
        {
            InitializeComponent();

            labelName = LoadLabelName();
            num_output_nodes = labelName.Length;

            nn = new NeuralNetwork();
            nn.AddLayer(num_input_nodes);
            nn.AddLayer(num_hidden_nodes);
            nn.AddLayer(num_output_nodes);


            // アプリ開始時に手描きデータを読み込み
            LoadData();

            // ランダムでラベルを表示する
            setText2((pixel.Count + 1).ToString());
            currentLabelIndex = random.Next(labelName.Length);
            setText1(labelName[currentLabelIndex]);
        }

        // アプリ終了時に手描きデータを保存
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (modifyFlag)
            {
                DialogResult result = MessageBox.Show("手描きデータを保存しますか？", "Save", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (result == DialogResult.Yes)
                {
                    SaveData();
                }
            }
        }

        private void pictureBox1_MouseMove(object sender, MouseEventArgs e)
        {
            // 左ボタンが押されていたらマウス軌跡描画
            if ((Control.MouseButtons & MouseButtons.Left) == MouseButtons.Left)
            {
                Point pos = pictureBox1.PointToClient(Cursor.Position);
                PenPlot(pos.X, pos.Y);
            }
        }

        // Training
        private void button1_Click(object sender, EventArgs e)
        {
            Thread t = new Thread(new ThreadStart(train));
            t.Start();
            buttonEnable(false);
        }

        // Clear
        private void button2_Click(object sender, EventArgs e)
        {
            // キャンバスクリア
            currentPixel = new double[num_input_nodes];
            GraphInit();
        }

        // Next
        private void button3_Click(object sender, EventArgs e)
        {
            // データ追加
            pixel.Add(currentPixel);
            // ラベル追加
            labelIndex.Add(currentLabelIndex);
            LabelDataAdd(currentLabelIndex);
            modifyFlag = true;

            // キャンバスクリア
            currentPixel = new double[num_input_nodes];
            GraphInit();

            // ランダムでラベルを表示する
            setText2((pixel.Count + 1).ToString());
            currentLabelIndex = random.Next(labelName.Length);
            setText1(labelName[currentLabelIndex]);
        }

        // ボタンの有効化・無効化
        delegate void buttonEnableDelegate(bool b);
        void buttonEnable(bool b)
        {
            if (InvokeRequired)
            {
                Invoke(new buttonEnableDelegate(buttonEnable), new object[] { b });
                return;
            }
            button1.Enabled = b;
        }

        // テキスト表示
        delegate void setTextDelegate(string s);
        void setText1(string s)
        {
            if (InvokeRequired)
                Invoke(new setTextDelegate(setText1), new object[] { s });
            else
                label1.Text = s;
        }

        void setText2(string s)
        {
            if (InvokeRequired)
                Invoke(new setTextDelegate(setText2), new object[] { s });
            else
                label2.Text = s;
        }

        // トレーニング
        private void train()
        {
            setText1("");

            GraphInit();
            nn.InitWeight();

            num_training_data = pixel.Count * 8 / 10;
            num_test_data = pixel.Count - num_training_data;

            // データが1000件に満たない場合は同じデータで学習を繰り返す
            int training_count = Math.Max(1000, num_training_data);
            for (int i = 0; i < training_count; i++)
            {
                int j = i % num_training_data;

                nn.CalcForward(pixel[j]);
                nn.CalcError(label[j]);
                nn.UpdateWeight(alpha);

                // 100件ごとに精度検証と表示
                if (i % 100 == 0)
                {
                    Test(num_test_data, (double)i / (double)training_count);
                    setText2(i.ToString());
                }
            }

            // 最終精度表示
            setText2("Testing...");
            double score = Test(num_test_data, 1.0);
            setText2((score * 100.0).ToString("F2") + "%");

            // 結果の保存
            DialogResult result = MessageBox.Show("学習結果を保存しますか？", "Save", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
            if (result == DialogResult.Yes)
            {
                nn.SaveWeight();
                SaveLabelName(labelName);
            }

            buttonEnable(true);
        }

        // 精度検証
        double Test(int dataLength, double graphX)
        {
            int ok = 0;
            int offset = num_training_data;
            int lastData = offset + dataLength;
            for (int i = 0; i < dataLength; i++)
            {
                nn.CalcForward(pixel[offset + i]);
                if (nn.GetMaxOutput() == labelIndex[offset + i])
                {
                    ok++;
                }
            }
            double score = (double)ok / (double)dataLength;
            GraphPlot(graphX, score);
            return score;
        }


        // これまで入力したピクセルデータとラベルデータがもしあれば読み込み
        public void LoadData()
        {
            if (!System.IO.File.Exists(@"pixel.dat"))
            {
                return;
            }

            using (FileStream fsPixel = new FileStream(@"pixel.dat", FileMode.Open))
            {
                using (BinaryReader brPixel = new BinaryReader(fsPixel))
                {
                    int num_data = brPixel.ReadInt32();
                    for (int i = 0; i < num_data; i++)
                    {
                        double[] data = new double[num_input_nodes];
                        for (int j = 0; j < num_input_nodes; j++)
                        {
                            data[j] = (double)brPixel.ReadByte() / 255.0 * 0.99 + 0.01;
                        }
                        pixel.Add(data);
                    }
                }
            }

            using (FileStream fsLabel = new FileStream(@"label.dat", FileMode.Open))
            {
                using (BinaryReader brLabel = new BinaryReader(fsLabel))
                {
                    int num_data = brLabel.ReadInt32();
                    for (int i = 0; i < num_data; i++)
                    {
                        int index = brLabel.ReadByte();
                        labelIndex.Add(index);
                        LabelDataAdd(index);
                    }
                }
            }
        }

        // ピクセルデータとラベルデータ保存
        public void SaveData()
        {
            using (FileStream fsPixel = new FileStream(@"pixel.dat", FileMode.OpenOrCreate))
            {
                using (BinaryWriter bwPixel = new BinaryWriter(fsPixel))
                {
                    int num_data = pixel.Count;
                    bwPixel.Write(num_data);
                    for (int i = 0; i < num_data; i++)
                    {
                        for (int j = 0; j < num_input_nodes; j++)
                        {
                            if (pixel[i][j] > 0.5)
                            {
                                bwPixel.Write((byte)255);
                            }
                            else
                            {
                                bwPixel.Write((byte)0);
                            }
                        }
                    }
                }
            }

            using (FileStream fsLabel = new FileStream(@"label.dat", FileMode.OpenOrCreate))
            {
                using (BinaryWriter bwLabel = new BinaryWriter(fsLabel))
                {
                    int num_data = labelIndex.Count;
                    bwLabel.Write(num_data);
                    for (int i = 0; i < num_data; i++)
                    {
                        bwLabel.Write((byte)labelIndex[i]);
                    }
                }
            }
        }

        void LabelDataAdd(int index)
        {
            double[] currentLabel = new double[num_output_nodes];
            for (int i = 0; i < num_output_nodes; i++)
            {
                currentLabel[i] = 0.01;
            }
            currentLabel[index] = 0.99;
            label.Add(currentLabel);
        }

        // ラベル名CSV読み込み
        private string[] LoadLabelName()
        {
            string[] names;

            using (StreamReader reader = new StreamReader(@"label.txt"))
            {
                string line = reader.ReadLine();
                names = line.Split(',');
            }

            return names;
        }

        // ラベル名CSV書き込み
        private void SaveLabelName(string[] labelNames)
        {
            using (StreamWriter writer = new StreamWriter(@"label.txt"))
            {
                for (int i = 0; i < labelNames.Length; i++)
                {
                    if (i > 0)
                        writer.Write(",");
                    writer.Write(labelNames[i]);
                }
            }
        }

        private void GraphInit()
        {
            px0 = -1;
            py0 = -1;
            Graphics graphic = pictureBox1.CreateGraphics();
            graphic.Clear(Color.White);
            graphic.Dispose();
        }

        private void PenPlot(int x, int y)
        {
            if ((x < 0) || x >= pictureBox1.Width) return;
            if ((y < 0) || y >= pictureBox1.Height) return;

            Graphics graphic = pictureBox1.CreateGraphics();
            graphic.FillEllipse(Brushes.Black, x, y, 25, 25);
            graphic.Dispose();

            int px = x * 28 / pictureBox1.Width;
            int py = y * 28 / pictureBox1.Height;

            currentPixel[py * 28 + px] = 1.0;
        }

        private void GraphPlot(double x, double y)
        {
            if (px0 < 0)
            {
                px0 = 0;
                py0 = pictureBox1.Height;
            }

            x = Math.Min(1.0, Math.Max(0.0, x));
            x = pictureBox1.Width * x;
            y = Math.Min(1.0, Math.Max(0.0, y));
            y = pictureBox1.Height * (1.0 - y);
            int px1 = (int)x;
            int py1 = (int)y;

            Graphics g = pictureBox1.CreateGraphics();
            g.DrawLine(Pens.Blue, px0, py0, px1, py1);
            px0 = px1;
            py0 = py1;
            g.Dispose();
        }
    }
}
