# （a）必要なライブラリのインポート
from keras.models import Sequential, Model
from keras.layers.convolutional import Conv2D
from keras.layers import BatchNormalization
from keras.layers.advanced_activations import LeakyReLU
from keras.layers import Flatten, Dense, Input, Reshape, Input
from keras.optimizers import Adam
from keras.layers.convolutional import Conv2DTranspose
from keras.preprocessing.image import ImageDataGenerator
import numpy as np
import glob, os, sys
import cv2
import datetime as dt
import pickle as pkl
from keras.utils.vis_utils import plot_model
import matplotlib.pyplot as plt 

REAL_IMAGE_PATH = '' # 本物画像のフォルダー
FAKE_IMAGE_PATH = './images' # 生成画像の保存先
BATCH_SIZE = 100
ITERATION_MAX = 30000 # イテレーションの上限
LR = 0.0002
ALPHA = 0.2
BETA_1 = 0.5

# （b）Generator（生成モデル）を作る関数の定義
# 生成画像は32×32ピクセルのサイズ
def Generator():
  # SequentialモデルでGeneratorを定義
  model = Sequential()

  # 1層目：入力層
  model.add(Dense(4*4*512, input_shape=(100, )))
  model.add(Reshape((4, 4, 512)))
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha=ALPHA))

  # 2層目：逆畳み込み層1
  model.add(
    Conv2DTranspose(
      filters = 256,
      kernel_size = 5,
      strides = 2,
      padding = 'same'
    )
  )
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha = ALPHA))

  # 3層目：逆畳み込み層2
  model.add(
    Conv2DTranspose(
      filters = 128,
      kernel_size = 5,
      strides = 2,
      padding = 'same'
    )
  )
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha = ALPHA))

  # 4層目：逆畳み込み層3
  model.add(
    Conv2DTranspose(
      filters = 3,
      kernel_size = 5,
      strides = 2,
      padding = 'same',
      activation = 'tanh'
    )
  )
  return model

# （c）Discriminator（識別モデル）を作る関数の定義
# 論文に従い、プーリング層は排除
def Discriminator():
  # SequentialモデルでDiscriminatorを定義
  model = Sequential()

  # 1層目
  model.add(
    Conv2D(
      filters = 32,
      kernel_size = 5,
      strides = 2,
      input_shape = (32, 32, 3),
      padding = 'same',
    )
  )
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha = ALPHA))

  # 2層目
  model.add(
    Conv2D(
      filters = 64,
      kernel_size = 5,
      strides = 2,
      padding = 'same',
    )
  )
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha = ALPHA))

  # 3層目
  model.add(
    Conv2D(
      filters = 128,
      kernel_size = 5,
      strides = 2,
      padding = 'same',
    )
  )
  model.add(BatchNormalization())
  model.add(LeakyReLU(alpha = ALPHA))
  model.add(Flatten())
  model.add(Dense(1, activation = 'sigmoid'))

  return model

# GeneratorとDiscriminatorを連結して
# Generatorトレーニング用のモデルを定義
def Combined(generator, discriminator):
  # Generatorの学習のみ行いたいので
  # Disctiminatorのパラメータ更新は行わない
  discriminator.trainable = False

  # GeneratorとDiscriminatorを結合
  model = Sequential([generator, discriminator])

  return model

# （d）本物画像を読み込む関数の定義
def loadRealImages():
  img_paths = []
  images = []

  imageFiles = glob.glob(os.path.join(REAL_IMAGE_PATH, '*.jpg'))

  for file in imageFiles:
    img = cv2.imread(file)
    img = cv2.resize(img, (32, 32))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    images.append(img)

  images = np.array(images)

  return np.array(images)

# （e）生成画像を保存する関数の定義
def saveGeneratImage(iteration, images):
  # Generatorで1度に25枚の画像を生成するため、表示エリアに5×5の
  # パネルを準備する
  fig, axes = plt.subplots(5, 5)
  images = generator.predict(noise)

  # 0 ～1にスケールを揃える
  images = 0.5 * images + 0.5

  # zipで表示する画像（save_gimage）と表示位置（axes.flatten）を
  # 対で取得し、順に表示（imshow）する
  for img, ax in zip(images, axes.flatten()):
    ax.imshow(img)
    ax.axis('off')

  if os.path.exists(FAKE_IMAGE_PATH) == False:
    os.mkdir(FAKE_IMAGE_PATH)
  fname = FAKE_IMAGE_PATH+'/generate_%05d.png' % iteration
  fig.savefig(fname)
  print()
  print('output: {}'.format(fname), end='', flush=True)
  save_gimage.append(images)

  plt.close()

if __name__ == '__main__':
  args = sys.argv 
    
  if len(args) == 1:
    print('生成画像のキーワードを指定してください。')
    sys.exit()

  REAL_IMAGE_PATH = args[1]
  # 本物画像のフォルダーをチェック。無ければ終了
  # 存在する場合はモデル構築、トレーニングを行う
  if os.path.exists(REAL_IMAGE_PATH) == False:
    print('本物画像のフォルダーがありません。')
    sys.exit()

  # （f）モデル構築と学習の実行
  # Generator単独でのCompileはしない。Combinedモデルに
  # 入れてからCompileを行う
  generator = Generator()

  # Discriminator
  discriminator = Discriminator()
  discriminator.compile(
    # 損失関数は2値交差エントロピー
    loss = 'binary_crossentropy',
    # 最適化アルゴリズムとしてAdamを指定
    optimizer = Adam(lr=LR, beta_1=BETA_1),
    # 評価関数
    metrics = ['accuracy']
  )

  # GeneratorとDisciminatorを連結したモデル
  combined = Combined(generator, discriminator)
  combined.compile(
    # 損失関数は2値交差エントロピー
    loss = 'binary_crossentropy',
    # 最適化アルゴリズムとしてAdamを指定
    optimizer = Adam(lr=LR, beta_1=BETA_1)
  )

  # 処理開始時刻の取得
  tstamp_s = dt.datetime.now().strftime("%H:%M:%S")

  checkPointNoise = np.random.uniform(-1, 1, (25, 100))
  checkPoint = 1000

  # 本物画像
  realImages = loadRealImages()

  # 本物画像を正規化 (0 to 255 >> -1 to 1)
  realImages = (realImages.astype(np.float32) - 127.5) / 127.5

  # 途中経過を保存する変数の定義
  save_gimage = []
  save_loss = []
  try:
    # 1回のイテレーションで100個の画像を取ってくることにする
    for iteration in range(ITERATION_MAX):
      # Discriminatorトレーニング
      idx = np.random.randint(0, len(realImages), BATCH_SIZE)
      # 本物画像の取得
      rimage = realImages[idx]
      noise = np.random.uniform(-1, 1, size=(BATCH_SIZE, 100))
      gimage = generator.predict(noise)
      rimage = rimage.reshape(BATCH_SIZE, 32, 32, 3)
      d_loss_real = discriminator.train_on_batch(rimage, np.ones((BATCH_SIZE, 1)))
      d_loss_fake = discriminator.train_on_batch(gimage, np.zeros((BATCH_SIZE, 1)))
      d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)
      noise = np.random.uniform(-1, 1, (BATCH_SIZE, 100))
      g_loss = combined.train_on_batch(noise, np.ones((BATCH_SIZE, 1)))
      
      # 1000イテレーションごとに生成画像を出力
      if iteration % checkPoint == 0:
        saveGeneratImage(iteration, checkPointNoise)

      # 10イテレーションごとに損失関数の値を表示
      if iteration % 10 == 0:
        print('')
        print('{0} Iteration={1}/{2}, DLoss={3:.4F}, GLoss={4:.4F}'.format(
          dt.datetime.now().strftime("%H:%M:%S"), iteration,
          ITERATION_MAX, d_loss[0], g_loss), end='', flush=True)
        save_loss.append((d_loss[0], g_loss))
      else:
        print('.', end='', flush=True)

    # 学習完了後のモデルで画像を生成する
    saveGeneratImage(iteration, checkPointNoise)
    # pkl形式でも生成画像を保存
    with open('save_dcgan_image.pkl', 'wb') as f:
      pkl.dump(save_gimage, f)
    # pkl形式で損失関数の値を保存
    with open('save_dcgan_loss.pkl', 'wb') as f:
      pkl.dump(save_loss, f)

    # 処理終了時刻の取得
    tstamp_e = dt.datetime.now().strftime("%H:%M:%S")
    time1 = dt.datetime.strptime(tstamp_s, "%H:%M:%S")
    time2 = dt.datetime.strptime(tstamp_e, "%H:%M:%S")

    # 処理時間を表示
    print('')
    print("開始: {0}、終了:{1}、処理時間:{2}".format(tstamp_s, tstamp_e, (time2 - time1)))

  except KeyboardInterrupt:
    print('')
    print('強制終了しました。')

    with open('save_dcgan_image.pkl', 'wb') as f:
      pkl.dump(save_gimage, f)

    with open('save_dcgan_loss.pkl', 'wb') as f:
      pkl.dump(save_loss, f)






