import chainer
import os
from chainer import functions as F
from chainer import links as L
from chainer.datasets import get_cifar10, SubDataset
from chainer.iterators import MultiprocessIterator
from chainer.training import StandardUpdater, Trainer
from chainer.optimizers import AdaGrad
from chainer.training.extensions import PrintReport,
  ProgressBar, LogReport, Evaluator, snapshot_object

class CNNClassifier(chainer.Chain):
  def __init__(self):
    super(CNNClassifier, self).__init__(
      conv1=L.Convolution2D(3,  24, 3), # (3, 32, 32) -> (24, 30, 30)
      conv2=L.Convolution2D(24, 36,  5), # (24, 15, 15) -> (48, 11, 11)
      conv3=L.Convolution2D(36, 18,  3), # (48, 9, 9) -> (18, 6, 6)
      fc4=L.Linear(162, 96), # 216 = 18 * 3 * 3
      fc5=L.Linear(96, 10),
    )
    self.train = True

  def __call__(self, x):
    # conv1->relu1->pool1->norm1
    h = F.max_pooling_2d(F.relu(self.conv1(x)),
      ksize=2, stride=2) # (24, 30, 30) -> (24, 15, 15)
    h = F.max_pooling_2d(F.relu(self.conv2(h)),
      ksize=3, stride=1) # (48, 11, 11) -> (48, 9, 9)
    h = F.max_pooling_2d(F.relu(self.conv3(h)),
      ksize=4, stride=2) # (18, 6, 6) -> (18, 3, 3)
    h = F.relu(self.fc4(h))
    h = self.fc5(h)
    return h

def main():
  model = CNNClassifier()
  classifier = L.Classifier(model)
  train, test = get_cifar10(withlabel=True, ndim=3, scale=1.0)
  train_iter = MultiprocessIterator(train, 32)
  test_iter = MultiprocessIterator(test, 32, repeat=False, shuffle=False)
  optimizer = AdaGrad(0.01)
  optimizer.setup(classifier)
  updater = StandardUpdater(train_iter, optimizer)
  trainer = Trainer(updater, (32, 'epoch'), out='./result')
  trainer.extend(Evaluator(test_iter, classifier))
  trainer.extend(LogReport())
  trainer.extend(PrintReport(
    ['epoch', 'main/loss', 'validation/main/loss',
      'main/accuracy', 'validation/main/accuracy', 'elapsed_time']))
  trainer.extend(ProgressBar(update_interval=1))
  trainer.extend(snapshot_object(model, 'model_iter_{.updater.iteration}'))
  trainer.run()

if __name__ == '__main__':
  main()